/**
 * load package
 */
import { configDotenv } from "dotenv"

import { watch, series, src, dest } from 'gulp'
import plumber from 'gulp-plumber'
import filter from 'gulp-filter'
import gulpIf from 'gulp-if'

import rename from 'gulp-rename'
import { glob } from 'glob'

import sharpOptimizeImages from 'gulp-sharp-optimize-images'

import pug from 'gulp-pug'

import * as dartSass from 'sass'
import gulpSass from 'gulp-sass'
const sass = gulpSass(dartSass)
import postcss from 'gulp-postcss'
import autoprefixer from 'autoprefixer'
import tailwindcss from 'tailwindcss'
import csso from 'postcss-csso'

import * as esbuild from 'esbuild'

import browserSync from 'browser-sync'
import gulpConnectPhp from 'gulp-connect-php'


/**
 * const variable
 */
const dotenvData = configDotenv({path: '.env'}).parsed ?? {}

const nodeEnv = (process.env.NODE_ENV ?? '').split('-')

const phpIni = dotenvData.PHP_INI_PATH ?? '';

const htmlMode = ['html', 'php'][dotenvData.IS_PHP.toString() === 'true' ? 1 : 0];


/**
 * 必要なURLやPATHを.envから取得
 * @requires PRODモードのパス書き換えを忘れないように
 */
let dirpath = (() => {
  let mode = 'development'
  let strictAbsPath = [dotenvData?.DOMAIN_DEV, dotenvData?.DOCROOT_DEV, dotenvData?.SUBDIR_DEV].join('')
  let rootAbsPath = [dotenvData?.DOCROOT_DEV, dotenvData?.SUBDIR_DEV].join('')
  let innerPublicPath = [dotenvData?.PUBLIC_DEV].join('')
  let outerPublicPath = [dotenvData?.DOMAIN_DEV, dotenvData?.DOCROOT_DEV, dotenvData?.SUBDIR_DEV, dotenvData?.PUBLIC_DEV].join('')
  let version = dotenvData?.VERSION ?? ''

  let isTerminate = false

  for(let i = 0, l = nodeEnv.length; i < l; i ++){
    switch(nodeEnv[i]){
      case 'local':
        mode = 'production-' + nodeEnv[i]
        strictAbsPath = [dotenvData?.DOMAIN_LOCAL, dotenvData?.DOCROOT_LOCAL, dotenvData?.SUBDIR_LOCAL].join('')
        rootAbsPath = [dotenvData?.DOCROOT_LOCAL, dotenvData?.SUBDIR_LOCAL].join('')
        innerPublicPath = [dotenvData?.PUBLIC_LOCAL].join('')
        outerPublicPath = [dotenvData?.DOMAIN_LOCAL, dotenvData?.DOCROOT_LOCAL, dotenvData?.SUBDIR_LOCAL, dotenvData?.PUBLIC_LOCAL].join('')
        isTerminate = true
      break

      case 'server':
        mode = 'production-' + nodeEnv[i]
        strictAbsPath = [dotenvData?.DOMAIN_SERVER, dotenvData?.DOCROOT_SERVER, dotenvData?.SUBDIR_SERVER].join('')
        rootAbsPath = [dotenvData?.DOCROOT_SERVER, dotenvData?.SUBDIR_SERVER].join('')
        innerPublicPath = [dotenvData?.PUBLIC_SERVER].join('')
        outerPublicPath = [dotenvData?.DOMAIN_SERVER, dotenvData?.DOCROOT_SERVER, dotenvData?.SUBDIR_SERVER, dotenvData?.PUBLIC_SERVER].join('')
        isTerminate = true
      break
    }

    if(isTerminate) break
  }

  return {
    mode: mode,
    strictAbsPath: strictAbsPath,
    rootAbsPath: rootAbsPath,
    innerPublicPath: innerPublicPath,
    outerPublicPath: outerPublicPath,
    version: version,
  }
})()


/**
 * task: copy
 * @listens src/public/*
 * @exports dist/*
 */
const task_copy = done => {
  src('./src/public/**/*', {
    allowEmpty: true,
    dot: true,
  })
  .pipe(plumber())
  .pipe(dest('./dist'))

  done()
}


/**
 * task: copy & minify
 * @listens src/img/*
 * @exports dist/img/*
 */
const task_img = done => {
  const png = filter('**/*.png', {restore: true})
  const jpg = filter('**/*.jpg', {restore: true})

  src('./src/img/**/*.*', {
    allowEmpty: true,
  })
  .pipe(plumber())

  .pipe(png)
  .pipe(sharpOptimizeImages({
    png_to_webp: {
      quality: 80,
      lossless: false,
    },
    png: {},
  }))
  .pipe(png.restore)

  .pipe(jpg)
  .pipe(sharpOptimizeImages({
    jpg_to_webp: {
      quality: 80,
      lossless: false,
    },
    jpg: {
      quality: 80,
      mozjpeg: true,
    },
  }))
  .pipe(jpg.restore)

  .pipe(dest('./dist/img'))

  done()
}


/**
 * task: css
 * @listens src/css/*
 * @exports dist/css/*
 */
const task_css = done => {
  src('./src/**/!(_)*.scss', {
    arrowEmpty: true,
    sourcemaps: true,
  })
  .pipe(plumber())
  .pipe(sass({
    outputStyle: 'expanded',
  }))
  .pipe(postcss([
    autoprefixer(),
    tailwindcss(),
    csso(),
  ]))
  .pipe(dest('./dist', {
    sourcemaps: '.',
  }))

  done()
}


/**
 * task: js
 * @listens src/js/*
 * @exports dist/js/*
 */
const task_js = async done => {
  let files = await glob('./src/js/**/!(_)*.{js,ts}', {
    ignore: 'node_modules/**'
  })

  await esbuild.build({
    entryPoints: files,
    outdir: 'dist/js',
    target: ['es6'],
    bundle: true,
    minify: true,
    sourcemap: true,
    logLevel: 'info',
  })

  done()
}


/**
 * task: html
 * @listens src/*
 * @exports dist/*
 */
const task_html = done => {
  src('./src/**/!(_)*.pug', {
    allowEmpty: true,
  })
  .pipe(plumber())
  .pipe(pug({
    locals: {
      mode: dirpath.mode,
      strictAbsPath: dirpath.strictAbsPath,
      rootAbsPath: dirpath.rootAbsPath,
      innerPublicPath: dirpath.innerPublicPath,
      outerPublicPath: dirpath.outerPublicPath,
      version: dirpath.version,
    },
    filters: {
      "php": text => "<?php\r\n" + text + "\r\n?>"
    },
  }))
  .pipe(gulpIf(htmlMode === 'php', rename({
    extname: '.php',
  })))
  .pipe(dest('./dist'))

  done()
}


/**
 * task: watch
 */
const task_watch = done => {
  watch('./src/public/**/*', series(task_copy, task_reload))
  watch('./src/img/**/*', series(task_img, task_reload))
  watch('./src/**/*.pug', series(task_html, task_reload))
  watch('./src/**/*.scss', series(task_css, task_reload))
  watch('./src/**/*.{js,ts}', series(task_js, task_reload))
  done()
}

const task_server = done => {
  if(htmlMode === 'html'){
    browserSync.init({
      server: {
        baseDir: './dist',
      },
    })
  }else{
    let options = {
      base: './dist',
      port: 3000,
    }

    if(phpIni !== ''){
      options.ini = phpIni;
    }

    gulpConnectPhp.server(options,
    () => {
      browserSync.init({
        proxy: 'localhost:3000',
      })
    })
  }

  done()
}

const task_reload = done => {
  browserSync.reload()
  done()
}


/**
 * exports
 */
export default series(
  task_copy,
  task_img,
  task_css,
  task_html,
  task_js,
)

export const dev = series(
  task_copy,
  task_img,
  task_css,
  task_html,
  task_js,
  task_server,
  task_watch,
)
